/* Automatic Fan Switch for Power Supply
 * run fan if current is above some settable threshold
 * run fan if temperature is above some settable threshold
 * Don't make fan state changes any faster than 30 seconds apart
 * display current temp and current current (ha!)
 * Two button UI allows changing of all the thresholds.
 * Store threshold values in EEPROM.
 * There is both an upper and lower threshold to give each decision
 * some hysteresis.
 *
 * William F. Dudley Jr.
 * 2021 11 14
 * public domain
 */
#include <Arduino.h>
#include <EEPROM.h>
#include "EEPROM_rw_anything.h"
#define DS18B20 1
#if DS18B20
#include <OneWire.h> 
#define DEBUG_DS18B20 0

// Note: Arduino Nano A6 and A7 cannot be used as digital pins

#define BREADBOARD_VERSION 0
#if BREADBOARD_VERSION
#define DS18B20_PIN 3
#define SEG_A 10
#define SEG_B 13
#define SEG_C 8
#define SEG_D 7
#define SEG_E 6
#define SEG_F 12
#define SEG_G 11
#define SEGDP 9
#define CATH_1 4
#define CATH_2 5
#define ACS725  A0
#define RELAY   A1
#define BUTTON1 A2
#define BUTTON2 A3
#else
#define DS18B20_PIN 12
#define SEG_A 6
#define SEG_B 4
#define SEG_C A5
#define SEG_D A3
#define SEG_E A2
#define SEG_F 7
#define SEG_G 5
#define SEGDP 10
#define CATH_1 A4
#define CATH_2 3
#define ACS725  A7
#define RELAY   2
#define BUTTON1 8
#define BUTTON2 9
#endif

OneWire ds(DS18B20_PIN);

byte ds_addr[8];
byte type_s;
unsigned long start_ds_time;
unsigned long next_ds_time;
float current_threshold_high;
float current_threshold_low;
float temp_c_threshold_high;
float temp_c_threshold_low;
unsigned long last_relay_change_time;
unsigned long last_time_fan_is_running;
unsigned long toggle_display_time;
uint8_t fan_request, last_fan_request, relay_state, last_relay_state, first_run, disp_mode;


void find_ds(void) {
byte i;
  memset(ds_addr, '\0', sizeof(ds_addr));
  if ( !ds.search(ds_addr)) {
    Serial.println(F("No more addresses."));
    Serial.println();
    ds.reset_search();
    delay(250);
    return;
  }

  Serial.print(F("ROM ="));
  for( i = 0; i < 8; i++) {
    Serial.write(' ');
    Serial.print(ds_addr[i], HEX);
  }

  if (OneWire::crc8(ds_addr, 7) != ds_addr[7]) {
      Serial.println(F("CRC is not valid!"));
      return;
  }
  Serial.println();
}

void id_ds_type (void) {
  // the first ROM byte indicates which chip
  switch (ds_addr[0]) {
    case 0x10:
      Serial.println(F("Chip = DS18S20"));  // or old DS1820
      type_s = 1;
      break;
    case 0x28:
      Serial.println(F("Chip = DS18B20"));
      type_s = 0;
      break;
    case 0x22:
      Serial.println(F("Chip = DS1822"));
      type_s = 0;
      break;
    default:
      Serial.println(F("Device is not a DS18x20 family device."));
      return;
  } 
}

void start_ds(void) {
  ds.reset();
  ds.select(ds_addr);
  ds.write(0x44, 1);        // start conversion, with parasite power on at the end
  
}

float read_ds(void) {
  byte i;
  byte present = 0;
  byte data[12];
  float celsius, fahrenheit;

  // we might do a ds.depower() here, but the reset will take care of it.
  
  present = ds.reset();
  ds.select(ds_addr);    
  ds.write(0xBE);         // Read Scratchpad

#if DEBUG_DS18B20  
  Serial.print("  Data = ");
  Serial.print(present, HEX);
  Serial.print(" ");
#endif
  for ( i = 0; i < 9; i++) {           // we need 9 bytes
    data[i] = ds.read();
#if DEBUG_DS18B20  
    Serial.print(data[i], HEX);
    Serial.print(" ");
#endif
  }
#if DEBUG_DS18B20  
  Serial.print(" CRC=");
  Serial.print(OneWire::crc8(data, 8), HEX);
  Serial.println();
#endif

  // Convert the data to actual temperature
  // because the result is a 16 bit signed integer, it should
  // be stored to an "int16_t" type, which is always 16 bits
  // even when compiled on a 32 bit processor.
  int16_t raw = (data[1] << 8) | data[0];
  if (type_s) {
    raw = raw << 3; // 9 bit resolution default
    if (data[7] == 0x10) {
      // "count remain" gives full 12 bit resolution
      raw = (raw & 0xFFF0) + 12 - data[6];
    }
  } else {
    byte cfg = (data[4] & 0x60);
    // at lower res, the low bits are undefined, so let's zero them
    if (cfg == 0x00) raw = raw & ~7;  // 9 bit resolution, 93.75 ms
    else if (cfg == 0x20) raw = raw & ~3; // 10 bit res, 187.5 ms
    else if (cfg == 0x40) raw = raw & ~1; // 11 bit res, 375 ms
    //// default is 12 bit resolution, 750 ms conversion time
  }
  celsius = (float)raw / 16.0;
  // fahrenheit = celsius * 1.8 + 32.0;
#if DEBUG_DS18B20  
  Serial.print(F("Temperature = "));
  Serial.print(celsius);
  Serial.println(" Celsius, ");
  // Serial.print(fahrenheit);
  // Serial.println(" Fahrenheit");
#endif
  return celsius;
}
#endif

#define TM1637_LED_DISPLAY 0
#define BARE_LED_DISPLAY 1

#if TM1637_LED_DISPLAY
#include "TM1637.h"

TM1637 TM;

#define TM_CLK  2
#define TM_DATA 3
#endif

typedef struct config
{
    long magic_number;
    int16_t i_current_threshold_high;
    int16_t i_current_threshold_low;
    int16_t i_temp_c_threshold_high;
    int16_t i_temp_c_threshold_low;
} CONFIG_DATA;

CONFIG_DATA config_data_copy;

#define CONFIG_REVISION 12348L

// time in seconds
#define MINIMUM_FAN_DURATION 30

#if BARE_LED_DISPLAY
// uint8_t seg_pins[] = { 10, 13, 8, 7, 6, 12, 11, 9 };
// uint8_t cathode_pins[] = { 4, 5 };
uint8_t seg_pins[] = { SEG_A, SEG_B, SEG_C, SEG_D, SEG_E, SEG_F, SEG_G, SEGDP };
uint8_t cathode_pins[] = { CATH_1, CATH_2 };
#define NUM_DIGITS 2
uint8_t dbuf[NUM_DIGITS];
uint8_t disp_position;

/***************
   ---
  |   |
   ---
  |   |
   --- .

      -01-
  20 |    | 02
      -40-
  10 |    | 04
      -08-  .80
*/

static uint8_t seg[] =
{
  0x3f, 0x06, 0x5b, 0x4f, 0x66, 0x6d, 0x7d, 0x07, 0x7f, 0x6f,   // 0 - 9
  0x77, 0x7c, 0x39, 0x5e, 0x79, 0x71, 0x00, 0x40                // A - F, ' ', '-'
};

void writeByte(uint8_t segments) {
  for(uint8_t i = 0 ; i < 8 ; i++) {
    digitalWrite(seg_pins[i], segments & 1);
    segments >>= 1;
  }
}

void showDigit(uint8_t dignum) {
uint8_t segments;
    digitalWrite(cathode_pins[dignum ^ 1], HIGH); // only good for 2 digits
    digitalWrite(cathode_pins[dignum], LOW);
    segments = seg[dbuf[dignum] & 0x1f];
    if(dbuf[dignum] & 0x80) { segments |= 0x80; }
    writeByte(segments);
}

void init_7seg_io (void) {
  for(uint8_t i = 0 ; i < 8 ; i++) {
    pinMode(seg_pins[i], OUTPUT);
  }
  for(uint8_t i = 0 ; i < NUM_DIGITS ; i++) {
    pinMode(cathode_pins[i], OUTPUT);
  }
}
#endif

#ifdef BUTTON1
unsigned long last_B1_change_time;
unsigned long last_B2_change_time;
unsigned long last_B_change_time;
uint8_t ui_mode;
#define UI_IDLE 0
#define SHOW_HIGH_CURR_LIMIT 1
#define SHOW_LOW_CURR_LIMIT 2
#define SHOW_HIGH_TEMP_LIMIT 3
#define SHOW_LOW_TEMP_LIMIT 4

uint8_t button_scan(void) {
uint8_t r=0, b1, b2;
static uint8_t lb1, lb2;
    if(millis() - last_B1_change_time > 250) {
	b1 = digitalRead(BUTTON1);
	if(!b1) {
	    last_B1_change_time = millis();
	    // Serial.println(F("button 1"));
	    r |= 1;
	}
    }
    if(millis() - last_B2_change_time > 250) {
	b2 = digitalRead(BUTTON2);
	if(!b2) {
	    last_B2_change_time = millis();
	    // Serial.println(F("button 2"));
	    r |= 2;
	}
    }
    // if(r) {
	// Serial.print(F("button_scan "));
	// Serial.println(r);
    // }
    return r;
}
#endif

void setup(void)
{
    Serial.begin(115200);
    Serial.println(__FILE__);

#if TM1637_LED_DISPLAY
    TM.init(TM_CLK, TM_DATA);
    TM.setBrightness(1);
    TM.displayClear();
#endif
    pinMode(RELAY, OUTPUT);
#ifdef BUTTON1
    pinMode(BUTTON1, INPUT_PULLUP);
    pinMode(BUTTON2, INPUT_PULLUP);
    ui_mode = UI_IDLE;
#endif
    EEPROM_readAnything(0, config_data_copy); // get saved settings
    if(config_data_copy.magic_number != CONFIG_REVISION) {
	config_data_copy.magic_number = CONFIG_REVISION;
	config_data_copy.i_current_threshold_high = 10; // 1.0 amps
	config_data_copy.i_current_threshold_low = 8; // 0.8 amps
	config_data_copy.i_temp_c_threshold_high = 300; // 30.0 C
	config_data_copy.i_temp_c_threshold_low = 280; // 28.0 C
	EEPROM_writeAnything(0, config_data_copy);
	Serial.println(F("Initializing EEPROM with default values."));
    }
    current_threshold_high = ((float)config_data_copy.i_current_threshold_high) / 10.0 ;
    current_threshold_low = ((float)config_data_copy.i_current_threshold_low) / 10.0 ;
    temp_c_threshold_high = ((float)config_data_copy.i_temp_c_threshold_high) / 10.0 ;
    temp_c_threshold_low = ((float)config_data_copy.i_temp_c_threshold_low) / 10.0 ;
    Serial.print(F("i_current_threshold_high = ")); Serial.println(config_data_copy.i_current_threshold_high);
    Serial.print(F("i_temp_c_threshold_high = ")); Serial.println(config_data_copy.i_temp_c_threshold_high);
    Serial.print(F("current_threshold_high = ")); Serial.println(current_threshold_high);
    Serial.print(F("temp_c_threshold_high = ")); Serial.println(temp_c_threshold_high);
    relay_state = 0;
    fan_request = 0;
    last_fan_request = 2;	// illegal state
    last_relay_state = 2;	// illegal state
    last_relay_change_time = 0L;
    first_run = 1;
#if BARE_LED_DISPLAY
    init_7seg_io ();
    dbuf[0] = 0x80;
    dbuf[1] = 0;
#endif
#if DS18B20
    find_ds();
    id_ds_type();
#endif
}

void loop(void)
{
static float filt_current;
static uint16_t disp_count;
static uint8_t ds_has_started;
float disp_current;
float temp_C;
char buff[8];
uint8_t key, i, j, print_current;
    print_current = 0;
#if DS18B20
    if(millis() - next_ds_time > 5000) { // 5 sec
	start_ds();
	next_ds_time = millis();
	start_ds_time = millis();
	ds_has_started = 1;
    }
    if(ds_has_started && millis() - start_ds_time > 1000) { // 5 sec
	temp_C = read_ds();
	Serial.print(F("Temperature = "));
	Serial.print(temp_C);
	Serial.print(F(" C, high limit is "));
	Serial.print(temp_c_threshold_high);
	Serial.print(F(" C, low limit is "));
	Serial.println(temp_c_threshold_low);
	print_current++;
	ds_has_started = 0;
	if(temp_C > temp_c_threshold_high) { fan_request |= 1; }
	if(temp_C < temp_c_threshold_low) { fan_request &= ~1; }
	if(fan_request & 1 && !(last_fan_request & 1))
	{
	    Serial.println(F("fan on request for over high temp"));
	}
    }
#endif
#ifdef BUTTON1
  key = button_scan();
  if(key) {
	Serial.print(F("button_scan returns "));
	Serial.println(key);
      if(key & 1) {
	last_B_change_time = millis();
	switch(ui_mode) {
	    case UI_IDLE :
		ui_mode = SHOW_HIGH_CURR_LIMIT;
		Serial.println(F("ui_mode = SHOW_HIGH_CURR_LIMIT"));
		break;
	    case SHOW_HIGH_CURR_LIMIT :
		if(config_data_copy.i_current_threshold_high < 99) { // 9.9 A max
		    config_data_copy.i_current_threshold_high++; // 100mA change
  Serial.print(F("i_current_threshold_high = ")); Serial.println(config_data_copy.i_current_threshold_high);
		    current_threshold_high = ((float)config_data_copy.i_current_threshold_high) / 10.0 ;
  Serial.print(F("current_threshold_high = ")); Serial.println(current_threshold_high);
		}
		break;
	    case SHOW_LOW_CURR_LIMIT :
		if(config_data_copy.i_current_threshold_low < config_data_copy.i_current_threshold_high - 1) {
		    config_data_copy.i_current_threshold_low++; // 100mA change
  Serial.print(F("i_current_threshold_low = ")); Serial.println(config_data_copy.i_current_threshold_low);
		    current_threshold_low = ((float)config_data_copy.i_current_threshold_low) / 10.0 ;
  Serial.print(F("current_threshold_low = ")); Serial.println(current_threshold_low);
		}
		break;
	    case SHOW_HIGH_TEMP_LIMIT :
		if(config_data_copy.i_temp_c_threshold_high < 800) {
		    config_data_copy.i_temp_c_threshold_high += 10; // whole degree
  Serial.print(F("i_temp_c_threshold_high = ")); Serial.println(config_data_copy.i_temp_c_threshold_high);
		    temp_c_threshold_high = ((float)config_data_copy.i_temp_c_threshold_high) / 10.0 ;
  Serial.print(F("temp_c_threshold_high = ")); Serial.println(temp_c_threshold_high);
		}
		break;
	    case SHOW_LOW_TEMP_LIMIT :
		if(config_data_copy.i_temp_c_threshold_low < config_data_copy.i_temp_c_threshold_high) {
		    config_data_copy.i_temp_c_threshold_low += 10; // whole degree
  Serial.print(F("i_temp_c_threshold_low = ")); Serial.println(config_data_copy.i_temp_c_threshold_low);
		    temp_c_threshold_low = ((float)config_data_copy.i_temp_c_threshold_low) / 10.0 ;
  Serial.print(F("temp_c_threshold_low = ")); Serial.println(temp_c_threshold_low);
		}
		break;
	}
      }
      if(key & 2) {
	last_B_change_time = millis();
	switch(ui_mode) {
	    case UI_IDLE :
		ui_mode = SHOW_HIGH_TEMP_LIMIT;
		Serial.println(F("ui_mode = SHOW_HIGH_TEMP_LIMIT"));
		break;
	    case SHOW_HIGH_CURR_LIMIT :
		if(config_data_copy.i_current_threshold_high > config_data_copy.i_current_threshold_low + 2) {
		    config_data_copy.i_current_threshold_high--; // 100mA change
  Serial.print(F("i_current_threshold_high = ")); Serial.println(config_data_copy.i_current_threshold_high);
		    current_threshold_high = ((float)config_data_copy.i_current_threshold_high) / 10.0 ;
  Serial.print(F("current_threshold_high = ")); Serial.println(current_threshold_high);
		}
		break;
	    case SHOW_LOW_CURR_LIMIT :
		if(config_data_copy.i_current_threshold_low > 1) {
		    config_data_copy.i_current_threshold_low--; // 100mA change
  Serial.print(F("i_current_threshold_low = ")); Serial.println(config_data_copy.i_current_threshold_low);
		    current_threshold_low = ((float)config_data_copy.i_current_threshold_low) / 10.0 ;
  Serial.print(F("current_threshold_low = ")); Serial.println(current_threshold_low);
		}
		break;
	    case SHOW_HIGH_TEMP_LIMIT :
		if(config_data_copy.i_temp_c_threshold_high > config_data_copy.i_temp_c_threshold_low + 10) {
		    config_data_copy.i_temp_c_threshold_high -= 10;// whole degree
  Serial.print(F("i_temp_c_threshold_high = ")); Serial.println(config_data_copy.i_temp_c_threshold_high);
		    temp_c_threshold_high = ((float)config_data_copy.i_temp_c_threshold_high) / 10.0 ;
  Serial.print(F("temp_c_threshold_high = ")); Serial.println(temp_c_threshold_high);
		}
		break;
	    case SHOW_LOW_TEMP_LIMIT :
		if(config_data_copy.i_temp_c_threshold_low > 200) {
		    config_data_copy.i_temp_c_threshold_low -= 10;// whole degree
  Serial.print(F("i_temp_c_threshold_low = ")); Serial.println(config_data_copy.i_temp_c_threshold_low);
		    temp_c_threshold_low = ((float)config_data_copy.i_temp_c_threshold_low) / 10.0 ;
  Serial.print(F("temp_c_threshold_low = ")); Serial.println(temp_c_threshold_low);
		}
		break;
	}
      }
  }
  else {
      if(millis() - last_B_change_time > 5000) { // 5 sec, no button press
        switch(ui_mode) {
	  case SHOW_HIGH_CURR_LIMIT :
	    ui_mode = SHOW_LOW_CURR_LIMIT;
	    Serial.println(F("ui_mode = SHOW_LOW_CURR_LIMIT"));
	    last_B_change_time = millis();
	    break;
	  case SHOW_HIGH_TEMP_LIMIT :
	    ui_mode = SHOW_LOW_TEMP_LIMIT;
	    Serial.println(F("ui_mode = SHOW_LOW_TEMP_LIMIT"));
	    last_B_change_time = millis();
	    break;
	  case SHOW_LOW_CURR_LIMIT :
	  case SHOW_LOW_TEMP_LIMIT :
	    ui_mode = UI_IDLE;
	    Serial.println(F("ui_mode = UI_IDLE"));
	    CONFIG_DATA config_data_copy2;
	    EEPROM_readAnything(0, config_data_copy2);
	    if(memcmp((void *)&config_data_copy, (void *)&config_data_copy2, sizeof(config_data_copy)/sizeof(char))) {
		EEPROM_writeAnything(0, config_data_copy);
	    }
	    break;
	}
      }
  }
#endif
  // Atmega 328 has 10 bit A/D; 5V implies that 1 count is .00488 mV
  // ACS725LLC-10AU output offset is Vcc/10, or about .33v, so 67 counts
  // Note: 67 seems low; actual offset fluctuates between 68 and 74.
  int16_t offset = 72;
  int16_t icurrent = analogRead(ACS725);
  int16_t raw_icurrent = icurrent;
  // ACS725LLC-10AU sensitivity is 264 mV/A
  // 1 count is 0.0184955 A
  // max value should be around 2.64V for 10A, or 540 counts
  // Serial.println(icurrent - offset);
  icurrent -= offset;
// #define CURRENT_SCALE 0.0184955
// experimentally determined:
#define CURRENT_SCALE 0.0169500
#define IIR_FILTER 1
#if IIR_FILTER
  if(icurrent < 0) { icurrent = 0; }
  // static int16_t loop_counter;
  // loop_counter++;
  // if(!(loop_counter % 400)) { Serial.println(F("400 !")); }
#endif
  double fcurrent = ((double)(icurrent)) * CURRENT_SCALE;
#if IIR_FILTER
  filt_current -= filt_current / 400.0;
  filt_current += fcurrent;
  disp_current = filt_current / 400.0;
#else
  disp_current = fcurrent;
#endif
  if(print_current) {
    Serial.print(F("Current = "));
    Serial.print(disp_current);
    Serial.print(F(" A, high limit is "));
    Serial.print(current_threshold_high);
    Serial.print(F(" A, low limit is "));
    Serial.print(current_threshold_low);
    Serial.print(F(" A, raw value is "));
    Serial.println(raw_icurrent);
  }
  if(disp_current > current_threshold_high) { fan_request |= 2; }
  if(disp_current < current_threshold_low) { fan_request &= ~2; }
  if(fan_request & 2 && !(last_fan_request & 2)) {
      Serial.println(F("fan on request for current threshold"));
  }
  relay_state = fan_request;
  if(relay_state) {
    last_time_fan_is_running = millis();
  }
  else if((millis() - last_time_fan_is_running) < (MINIMUM_FAN_DURATION * 1000L)) {
    relay_state = 1;
  }
  if(last_relay_state != relay_state
  && ((millis() - last_relay_change_time) > (MINIMUM_FAN_DURATION * 1000L)
      || first_run)) {
      digitalWrite(RELAY, relay_state);
      Serial.print(F("relay "));
      Serial.println(relay_state ? F("ON") : F("OFF"));
      last_relay_state = relay_state;
      last_relay_change_time = millis();
      first_run = 0;
      Serial.print(F("change at "));
      Serial.print(last_relay_change_time / 1000L);
      Serial.println(F(" seconds."));
  }
#if TM1637_LED_DISPLAY
  disp_count++;
  if(disp_count >= 500) {
      if(disp_current < 0.0) { disp_current = 0.0; }
      TM.displayFloat(disp_current);
      disp_count = 0;
  }
#endif
  delay(1);
  if(millis() - toggle_display_time > 4000) { // 4 sec, no button press
    disp_mode ^= 1;
    toggle_display_time = millis();
  }
#if BARE_LED_DISPLAY
  {
      float led_number;
      if(ui_mode == UI_IDLE) {
	  led_number = (disp_mode) ? (disp_current + 0.05) : (temp_C + 0.5);
	  if(print_current) {
	      // Serial.print(F("led_number ")); Serial.println(led_number);
	  }
      }
      else if(ui_mode == SHOW_HIGH_CURR_LIMIT) {
	  led_number = current_threshold_high + 0.05;
      }
      else if(ui_mode == SHOW_LOW_CURR_LIMIT) {
	  led_number = current_threshold_low + 0.05;
      }
      else if(ui_mode == SHOW_HIGH_TEMP_LIMIT) {
	  led_number = temp_c_threshold_high + 0.05;
      }
      else if(ui_mode == SHOW_LOW_TEMP_LIMIT) {
	  led_number = temp_c_threshold_low + 0.05;
      }
      if(led_number < 10.0) {
	  buff[0] = '0' + int(led_number);
	  buff[1] = '.';
	  buff[2] = '0' + int(10 * (led_number - int(led_number)));
      }
      else {
	  int ln = (int)led_number;
	  buff[0] = '0' + int(ln / 10);
	  buff[1] = '0' + int(ln % 10);
	  buff[2] = '.';
      }
      buff[3] = '\0';
      if(print_current) {
	  // Serial.print(F("led_number ")); Serial.println(buff);
      }
      for(i = 0, j = 0 ; i < 3 ; i++) {
	if(isdigit(buff[i])) {
	  dbuf[j] = buff[i] - '0';
	}
	else if(buff[i] == '.') {
	  if(j) {
	    dbuf[--j] |= 0x80;
	  }
	}
	j++;
      }
      if(print_current) {
	  // Serial.print(F("dbuf: ")); Serial.print(buff[0], HEX); Serial.print(" "); Serial.print(buff[1], HEX); Serial.print(" "); Serial.println(buff[2], HEX);
      }
      disp_position ^= 1;
      showDigit(disp_position);
  }
#endif
}

